# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116

import streamlink.webbrowser.cdp.devtools.browser as browser
import streamlink.webbrowser.cdp.devtools.debugger as debugger
import streamlink.webbrowser.cdp.devtools.dom as dom
import streamlink.webbrowser.cdp.devtools.emulation as emulation
import streamlink.webbrowser.cdp.devtools.fetch as fetch
import streamlink.webbrowser.cdp.devtools.input_ as input_
import streamlink.webbrowser.cdp.devtools.inspector as inspector
import streamlink.webbrowser.cdp.devtools.io as io
import streamlink.webbrowser.cdp.devtools.network as network
import streamlink.webbrowser.cdp.devtools.page as page
import streamlink.webbrowser.cdp.devtools.runtime as runtime
import streamlink.webbrowser.cdp.devtools.security as security
import streamlink.webbrowser.cdp.devtools.target as target
import streamlink.webbrowser.cdp.devtools.util as util
