"""
The entire API of the ``streamlink.webbrowser`` package is considered unstable and may change at any time.
Use at your own risk!
"""

from streamlink.webbrowser.chromium import ChromiumWebbrowser
from streamlink.webbrowser.exceptions import WebbrowserError
from streamlink.webbrowser.webbrowser import Webbrowser
